/*****************************************************
Chip type               : ATmega8
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/

#include <mega8.h>
#include <sleep.h>

#define A PORTC.2
#define B PORTC.3                 
#define C PORTC.4
#define D PORTC.5
#define E PORTD.0
#define F PORTD.1
#define G PORTD.3
#define DP PORTD.4
#define A0 PORTB.6
#define A1 PORTB.7
#define A2 PORTD.5
#define A3 PORTD.6

#define pwm_on do{TCCR1A|=1<<COM1A1|1<<COM1A0;}while(0)
#define pwm_off do{TCCR1A&=~(1<<COM1A1|1<<COM1A0);PORTB.1=1;}while(0)
#define Utop 225
#define relay PORTC.1
#define buzzer PORTC.0
#define button0 PINB.4

#define R_upr 10000 //Ohms
#define R_lwr 2000   //Ohms
#define U_max_adc ((256*(R_upr+R_lwr))/R_lwr) //=U_max*100,15.36V


volatile bit ovl_bit;
volatile unsigned char ADC7, utop;
volatile unsigned char tm0, tm1, tm2;

unsigned char display_buffer[4], d_point_position; 
volatile unsigned char display_counter;

enum btn_states {b_prs, b_rls, b_hold} btn0=b_rls;
enum dev_states {normal, started, undervoltage, overvoltage, lobat, sdn, sleep} 
dstate=sleep;

void ports_cfg()
{
PORTB=0b00010010;
DDRB=0b11101111;

PORTC=0b01000000;
DDRC=0b01111111;

PORTD=0b00000100;
DDRD=0b01111011;
}
void t1_cfg()
{
//fast PWM top - ICR1
TCCR1A=TCCR1B=TIMSK=TCNT1=0;
TCCR1A|=1<<WGM11;
TCCR1B|=1<<WGM13|1<<WGM12|1<<CS10;
OCR1A=0x0A;
ICR1=0xFF>>1;

}
void adc_cfg()
{
ADMUX=0; ADMUX|=1<<ADLAR|1<<REFS1|1<<REFS0|7;
ADCSRA=0; ADCSRA|=1<<ADEN|1<<ADSC|1<<ADIE|1<<ADPS2|0<<ADPS1|0<<ADPS0;// /16
}
void acmp_cfg()
{
 ACSR=0; ACSR|=1<<ACBG|1<<ACIE|0<<ACIS1|0<<ACIS0;
}
void eint_cfg()
{
GICR|=0x40;
MCUCR=0x03;
GIFR=0x40;
}

void scan_keys ()
{
static unsigned char count0,count1;

if (!button0) {count0++;} else {count0 = 0;count1 = 0;btn0 = b_rls;}
if ((count0 > 250)&&(btn0!=b_hold)) {btn0 = b_prs;count0 = 0;count1++;}
if ((count1 > 30)&&(btn0 == b_prs)) {btn0 = b_hold;count1 = 0;}
}

void clear_display()
{
 A0=A1=A2=A3=0; DP = 1;
}
void convert_data()
{
 unsigned char tmp;
 static char n;
 if(!tm0)
 {
  tm0=50; tmp=(long)ADC7*U_max_adc/(0xff*10); d_point_position=1;
  for(n=0;n<4;n++)
  {
   display_buffer[n]=tmp%10;tmp/=10;
  }
 }
}
void display_data ()
{
(display_counter>0)?(display_counter--):(display_counter = 4);
    clear_display();
  switch (display_buffer[display_counter])
 {
  case 0: A=0; B=0; C=0; D=0; E=0; F=0; G=1; break;
  case 1: A=1; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 2: A=0; B=0; C=1; D=0; E=0; F=1; G=0; break;
  case 3: A=0; B=0; C=0; D=0; E=1; F=1; G=0; break;
  case 4: A=1; B=0; C=0; D=1; E=1; F=0; G=0; break;
  case 5: A=0; B=1; C=0; D=0; E=1; F=0; G=0; break;
  case 6: A=0; B=1; C=0; D=0; E=0; F=0; G=0; break;
  case 7: A=0; B=0; C=0; D=1; E=1; F=1; G=1; break;
  case 8: A=0; B=0; C=0; D=0; E=0; F=0; G=0; break;
  case 9: A=0; B=0; C=0; D=0; E=1; F=0; G=0; break;
  case 10: A=1; B=1; C=1; D=1; E=1; F=1; G=1; break;
  case 11: A=1; B=1; C=0; D=0; E=0; F=1; G=1; break; //u
  case 12: A=1; B=1; C=1; D=0; E=0; F=1; G=0; break; //c
  case 13: A=1; B=1; C=0; D=0; E=0; F=0; G=0; break; //b
  case 14: A=1; B=1; C=1; D=0; E=0; F=0; G=0; break; //t
  case 15: A=1; B=1; C=1; D=0; E=0; F=1; G=1; break; //l
  case 16: A=1; B=1; C=0; D=0; E=0; F=1; G=0; break; //o
  case 17: A=0; B=0; C=0; D=1; E=0; F=0; G=0; break; //A
  case 18: A=1; B=0; C=0; D=1; E=0; F=0; G=0; break; //H
  case 19: A=0; B=1; C=1; D=0; E=0; F=0; G=1; break; //C
  case 20: A=0; B=0; C=1; D=1; E=0; F=0; G=0; break; //P
  case 21: A=0; B=1; C=1; D=0; E=0; F=0; G=0; break; //E
  case 22: A=1; B=0; C=0; D=0; E=0; F=0; G=1; break; //U 
  case 23: A=0; B=1; C=1; D=1; E=0; F=0; G=0; break; //F 
  case 24: A=1; B=1; C=1; D=0; E=0; F=0; G=1; break; //L
  case 25: A=1; B=1; C=1; D=1; E=0; F=1; G=0; break; //r
  case 26: A=1; B=1; C=1; D=0; E=1; F=1; G=1; break; //_
  case 27: A=1; B=0; C=0; D=0; E=0; F=1; G=0; break; //d 
  case 28: A=1; B=1; C=0; D=1; E=0; F=1; G=0; break; //n
  default: A=1; B=1; C=1; D=1; E=1; F=1; G=0; break; //- 
 }   
   
   switch (display_counter)
 {
  case 0: A0 = 1; if(d_point_position == 0) {DP = 0;} break;
  case 1: A1 = 1; if(d_point_position == 1) {DP = 0;} break;
  case 2: A2 = 1; if(d_point_position == 2) {DP = 0;} break;
  case 3: if(display_buffer[3]){A3 = 1;} if(d_point_position == 3) {DP = 0;} break;
  default: DP = 1; break;
 }
 }

void dev_mgmt();
char check_bat();

interrupt [EXT_INT0] void ext_int0_isr(void)
{
 pwm_off; utop=5; OCR1A=1; dstate=overvoltage;
}
interrupt [ANA_COMP] void ana_comp_isr(void)
{
 if(ACSR.ACO){pwm_on;}
 else {pwm_off;}      
}
interrupt [ADC_INT] void ADC_read ()
{
static unsigned char ccnt, n;
static unsigned int tmp;
    
    if (ccnt)
    {
    if ((OCR1AL>1)&&(ADCH>utop)) {OCR1AL--;ovl_bit=0;}
    else if ((OCR1AL<(ICR1>>1))&&(ADCH<utop)) {OCR1AL++;}
    else if(ADCH<(utop-15)) {ovl_bit=1;} 
    ccnt--; 
    if (!ccnt) {ADMUX=0b11100111;}
    }
    else 
    {
    ADMUX=0b11100110; ccnt=0xff; 
    if(n){tmp+=ADCH;n--;}
    else{n=32; ADC7=tmp>>5; tmp=0;}
    
    if(tm0)tm0--; 
    if(tm1)tm1--;
    if(tm2)tm2--;
    }
    ADCSR |= (1<<ADSC);
}


void main(void)
{
ports_cfg();
t1_cfg();
acmp_cfg();
adc_cfg();
eint_cfg();

utop=Utop;
dstate=normal; ADC7=240;

#pragma optsize-
WDTCR=0x1A;
WDTCR=0x0A;
#ifdef _OPTIMIZE_SIZE_
#pragma optsize+
#endif

#asm("sei")

while (1)
      {  
         scan_keys ();
         dev_mgmt();
         
       #asm("wdr")
      }
}

void dev_mgmt()    
{
 static bit bt;
 if(btn0==b_rls){bt=0;}

 switch (dstate)
 {
  
  case normal:
  {
   if ((btn0==b_hold)&&(!bt)) {dstate=sdn; bt=1;}
   convert_data();
   display_data();
   if(!check_bat()){dstate=lobat;}
   if(ovl_bit){dstate=undervoltage;}  
  }
  break; 
  
  case started:
  {
   pwm_on;utop=Utop;relay=1;
   sleep_disable(); 
   dstate=normal;  
  }
  break;
  
  case undervoltage:
  {
  static bit t;
  if(!ovl_bit) {dstate=normal;t=0;tm2=0;}  
  else if((!tm2)&&(!t)){tm2=250;t=1;} 
  else if((!tm2)&&(t)) {t=0; dstate=sdn;}
  else if (tm2<60) 
  {
   display_buffer[3]=24;
   display_buffer[2]=16;
   display_buffer[1]=10;
   display_buffer[0]=22;
   d_point_position=1;
   display_data();  
  }
  else {convert_data();display_data();}
  }
  break;  
  
  case overvoltage:
  {
  static bit n;  
  if((!tm2)&&(!n)){tm2=250;n=1;} 
  else if((!tm2)&&(n)) {n=0; dstate=sdn;}
  else  
  {
   display_buffer[3]=22;
   display_buffer[2]=21;
   display_buffer[1]=25;
   display_buffer[0]=25;
   d_point_position=3;
   display_data();  
  }
  }
  break;  
  
  case lobat:
  {
  static bit lbtr; 
  
  if((!tm1)&&(!lbtr)&&(!check_bat())) {tm1=250;lbtr=1;}
  if(check_bat()){tm1=0;lbtr=0;dstate=started;buzzer=0;break;} 
  if((!tm1)&&(lbtr)){dstate=sleep;lbtr=0;buzzer=0;break;}
  if(tm1)
   {
   display_buffer[3]=24;
   display_buffer[2]=13;
   display_buffer[1]=14;
   display_buffer[0]=tm1/50;
   d_point_position=1;
   display_data();
   buzzer=~buzzer;
   } 
  }
  break;
  
  case sdn:
  {
  pwm_off; relay=0; utop=5;
  clear_display(); 
  sleep_enable(); 
  idle(); 
  if (((btn0==b_prs)||(btn0==b_hold))&&(!bt)){dstate=started;bt=1;}
  }
  break;
  
  case sleep:
  {
  
  if(check_bat()){if(((btn0==b_prs)||(btn0==b_hold))&&(!bt))dstate=started;bt=1;} 
  else if ((btn0==b_prs)||(btn0==b_hold)){tm2=250;}
  else if (tm2) 
  {
   display_buffer[3]=24;
   display_buffer[2]=16;
   display_buffer[1]=13;
   display_buffer[0]=14;
   d_point_position=5;
   display_data();  
  } 
  else {dstate=sdn;}
  }
  break;

    
  default: break;
 }
}

char check_bat()
{
 static unsigned char tmp0,tmp1;
 static bit btrg; 
 if(tmp0!=ADC7)
 {
 tmp0=ADC7;
 tmp1=(unsigned long)tmp0*U_max_adc/(0xff*10); 
 } 
 if(tmp1>=120){btrg=1; return 1;} 
 
 if((tmp1>108)&&(btrg))return 1;
 else if(tmp1<=108){btrg=0;return 0;}
 else return 0;  
}